//
// sueda
// November, 2014
//

#include "Camera.h"
#include "MatrixStack.h"
#include <iostream>
#include <math.h>
#define GLM_FORCE_RADIANS
#include "glm/gtx/string_cast.hpp"
#include "glm/gtx/transform.hpp"

using namespace std;

Camera::Camera() :
   aspect(1.0f),
   fovy(30.0f/180.0f*M_PI),
   znear(0.1f),
   zfar(1000.0f),
   position(50.0f, 0.0f, 250.0f),
   direction(0.0f, 0.0f, 1.0f),
   yaw(0.0f),
   pitch(0.0f),
   window_width(WINDOW_SIZE_X),
   window_height(WINDOW_SIZE_Y),
   mouseInWindow(false),
   jumping(false)
{
}

Camera::~Camera()
{
   
}

void Camera::setWindowSize(float w, float h)
{
   aspect = w/h;
   window_width = w;
   window_height = h;
}

void Camera::update(const bool *keys, const glm::vec2 &mouse, vector<MapObject> map_objects)
{
   Contacts curCollision;
   MatrixStack OBJ;
   glm::mat4 camLocation;
   float dx = 0, dy = 0;
   int i;
   bool noCollisions = true;

   dx = mouse.x - window_width / 2;
   dy = mouse.y - window_height / 2;
   
   pitch -= MOUSE_PITCH_SENSITIVITY * dy;
   yaw -= MOUSE_YAW_SENSITIVITY * dx;

   if ((pitch / MOUSE_PITCH_SENSITIVITY) > MAX_PITCH) {
      pitch = MAX_PITCH * MOUSE_PITCH_SENSITIVITY;
   }
   else if ((pitch / MOUSE_PITCH_SENSITIVITY) < -MAX_PITCH) {
      pitch = -MAX_PITCH * MOUSE_PITCH_SENSITIVITY;
   }
   
   // Update camera position and orientation here
   glm::mat4 R = glm::rotate(yaw, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::vec3 right = glm::vec3(R[0][0], R[0][1], R[0][2]);
   glm::vec3 forward = -glm::vec3(R[2][0], R[2][1], R[2][2]);

   // forward (-z)
   if (keys['w'] && noCollisions) {
      position += MOVEMENT_SENSITIVITY * forward;

      OBJ.pushMatrix();
      OBJ.translate(position);
      camLocation = OBJ.topMatrix();
      OBJ.popMatrix();

      for (i = 0; i < map_objects.size() && noCollisions; i++) {
         curCollision = BoxCollision(camLocation, glm::vec3(8.0f, 8.0f, 8.0f),
          map_objects.at(i).getMatrix(), map_objects.at(i).getDimensions());
         
         if (curCollision.positions.size() != 0) {
            noCollisions = false;
            position -= MOVEMENT_SENSITIVITY * forward;
         }
      }
   }

   // backward (+z)
   if (keys['s'] && noCollisions) {
      position -= MOVEMENT_SENSITIVITY * forward;

      OBJ.pushMatrix();
      OBJ.translate(position);
      camLocation = OBJ.topMatrix();
      OBJ.popMatrix();

      for (i = 0; i < map_objects.size() && noCollisions; i++) {
         curCollision = BoxCollision(camLocation, glm::vec3(8.0f, 8.0f, 8.0f),
          map_objects.at(i).getMatrix(), map_objects.at(i).getDimensions());
         
         if (curCollision.positions.size() != 0) {
            noCollisions = false;
            position += MOVEMENT_SENSITIVITY * forward;
         }
      }
   }

   // left (-x)
   if (keys['a'] && noCollisions) {
      position -= MOVEMENT_SENSITIVITY * right;

      OBJ.pushMatrix();
      OBJ.translate(position);
      camLocation = OBJ.topMatrix();
      OBJ.popMatrix();

      for (i = 0; i < map_objects.size() && noCollisions; i++) {
         curCollision = BoxCollision(camLocation, glm::vec3(8.0f, 8.0f, 8.0f),
          map_objects.at(i).getMatrix(), map_objects.at(i).getDimensions());
         
         if (curCollision.positions.size() != 0) {
            noCollisions = false;
            position += MOVEMENT_SENSITIVITY * right;
         }
      }
   }

   // right (+x)
   if (keys['d'] && noCollisions) {
      position += MOVEMENT_SENSITIVITY * right;

      OBJ.pushMatrix();
      OBJ.translate(position);
      camLocation = OBJ.topMatrix();
      OBJ.popMatrix();

      for (i = 0; i < map_objects.size() && noCollisions; i++) {
         curCollision = BoxCollision(camLocation, glm::vec3(8.0f, 8.0f, 8.0f),
          map_objects.at(i).getMatrix(), map_objects.at(i).getDimensions());
         
         if (curCollision.positions.size() != 0) {
            noCollisions = false;
            position += MOVEMENT_SENSITIVITY * right;
         }
      }
   }
   
   // space (jump)
   if (keys[32]) {
      if (!jumping && position.y == 0.0f) {
         jumping = true;
      }
   }

   if (jumping) {
      if (position.y < JUMP_LIMIT) {
         position.y += 0.5f;
      }
      else {
         jumping = false;
      }
   }
   else {
      if (position.y > 0.0f) {
         position.y -= 0.5f;
      }
   }

   glutWarpPointer(window_width / 2, window_height / 2);

   glm::mat4 T = glm::translate(position);
   glm::mat4 RX = glm::rotate(pitch, glm::vec3(1.0f, 0.0f, 0.0f));
   glm::mat4 RY = glm::rotate(yaw, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 C = T * RY * RX;

   direction = -glm::vec3(C[2][0], C[2][1], C[2][2]);
}

void Camera::applyProjectionMatrix(MatrixStack *P) const
{
   P->perspective(fovy, aspect, znear, zfar);
}

void Camera::applyViewMatrix(MatrixStack *MV) const
{
   // Create the translation and rotation matrices
   glm::mat4 T = glm::translate(position);
   glm::mat4 RX = glm::rotate(pitch, glm::vec3(1.0f, 0.0f, 0.0f));
   glm::mat4 RY = glm::rotate(yaw, glm::vec3(0.0f, 1.0f, 0.0f));

   // The matrix C is the product of these matrices
   glm::mat4 C = T * RY * RX; // Also apply rotations here
   // C[3] location
   // C[2] direction
   // The view matrix is the inverse
   glm::mat4 V = glm::inverse(C);
   // Add to the matrix stack
   MV->multMatrix(V);
}
