//
// sueda
// November, 2014
//

#pragma  once
#ifndef __Camera__
#define __Camera__

#include <GL/freeglut.h>
#include "MatrixStack.h"
#include "MapObject.h"
#include "BoxCollision.h"
#include "glm/glm.hpp"

#define GLM_FORCE_RADIANS
#define MOVEMENT_SENSITIVITY 1.0f
#define MOUSE_PITCH_SENSITIVITY 0.0025f
#define MOUSE_YAW_SENSITIVITY 0.0025f
#define MAX_PITCH 360.0f
#define JUMP_LIMIT 7.0f
#define WINDOW_SIZE_X 600
#define WINDOW_SIZE_Y 600

class MatrixStack;

class Camera
{
   public:
      Camera();
      virtual ~Camera();
      void setWindowSize(float w, float h);
      void update(const bool *keys, const glm::vec2 &mouse, std::vector<MapObject> map_objects);
      void applyProjectionMatrix(MatrixStack *P) const;
      void applyViewMatrix(MatrixStack *MV) const;
      glm::vec3 getPosition() const { return position; }
      glm::vec3 getDirection() const { return direction; }
      float getWidth() const { return window_width; }
      float getHeight() const { return window_height; }
   
private:
   float aspect;
   float fovy;
   float znear;
   float zfar;
   float yaw;
   float pitch;
   float window_width;
   float window_height;
   bool mouseInWindow;
   bool jumping;
   glm::vec3 position;
   glm::vec3 direction;
};

#endif
