/**
 * Zach Mintzer
 * CPE471 Final Project
 */

#include <iostream>
#include <stdio.h>
#define GLM_FORCE_RADIANS
#include "glm/glm.hpp"
#include "TexturePlane.h"

TexturePlane::TexturePlane() :
   posBufID(0),
   texBufID(0),
   indBufID(0),
   textureID(0),
   tex_id(0)
{
}

TexturePlane::~TexturePlane()
{
}

void TexturePlane::load(std::vector<float> posBuf, std::vector<float> texBuf,
 std::vector<unsigned int> indBuf, glm::mat4 mat, glm::mat3 T)
{
   positionBuf = posBuf;
   texCoordsBuf = texBuf;
   indexBuf = indBuf;
   TransformationMat = mat;
   TextureTransform = T;
}

void TexturePlane::init()
{
   glGenBuffers(1, &posBufID);
   glBindBuffer(GL_ARRAY_BUFFER, posBufID);
   glBufferData(GL_ARRAY_BUFFER, positionBuf.size() * sizeof(float), 
    &positionBuf[0], GL_STATIC_DRAW);

   glGenBuffers(1, &texBufID);
   glBindBuffer(GL_ARRAY_BUFFER, texBufID);
   glBufferData(GL_ARRAY_BUFFER, texCoordsBuf.size() * sizeof(float), 
    &texCoordsBuf[0], GL_STATIC_DRAW);

   glGenBuffers(1, &indBufID);
   glBindBuffer(GL_ARRAY_BUFFER, indBufID);
   glBufferData(GL_ARRAY_BUFFER, indexBuf.size() * sizeof(unsigned int), 
    &indexBuf[0], GL_STATIC_DRAW);

   glBindBuffer(GL_ARRAY_BUFFER, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void TexturePlane::texInit(int id, const char *filename, int clamp)
{
   tex_id = id;
   
   Image *image = (Image *)malloc(sizeof(Image));
   if (image == NULL) {
      printf("Error allocating space for image\n");
   }
   if (!ImageLoad(filename, image)) {
      printf("Error loading texture image\n");
   }

   glActiveTexture(GL_TEXTURE0 + id);
   glGenTextures(1, &textureID);
   glBindTexture(GL_TEXTURE_2D, textureID);
   glTexImage2D(GL_TEXTURE_2D, 0, 3, image->sizeX, image->sizeY, 0, GL_RGB,
    GL_UNSIGNED_BYTE, image->data);
   glGenerateMipmap(GL_TEXTURE_2D);

   if (clamp) {
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
   }
   else {
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   }

   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
    GL_LINEAR_MIPMAP_LINEAR);
   glBindTexture(GL_TEXTURE_2D, 0);
   free(image);
}

void TexturePlane::draw(MatrixStack *MV, MatrixStack *P, GLint h_MV, GLint h_P, 
 GLint h_T, GLuint pid, GLint h_vertPos, GLint h_vertTexCoords, 
 GLint h_texture)
{
   glUseProgram(pid);
   glActiveTexture(GL_TEXTURE0 + tex_id);
   glBindTexture(GL_TEXTURE_2D, textureID);
   glUniform1i(h_texture, tex_id);
   
   glUniformMatrix4fv( h_P, 1, GL_FALSE, glm::value_ptr( P->topMatrix()));
   
   MV->pushMatrix();
   MV->multMatrix(TransformationMat);
   glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV->topMatrix()));
   MV->popMatrix();
   
   glUniformMatrix3fv(h_T, 1, GL_TRUE, glm::value_ptr(TextureTransform));
   
   // Enable and bind position array for drawing
   GLSL::enableVertexAttribArray(h_vertPos);
   glBindBuffer(GL_ARRAY_BUFFER, posBufID);
   glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);
   
   // Enable and bind texcoord array for drawing
   GLSL::enableVertexAttribArray(h_vertTexCoords);
   glBindBuffer(GL_ARRAY_BUFFER, texBufID);
   glVertexAttribPointer(h_vertTexCoords, 2, GL_FLOAT, GL_FALSE, 0, 0);
   
   // Bind index array for drawing
   int nIndices = (int)indexBuf.size();
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indBufID);
   
   // Draw
   glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
   
   // Disable and unbind
   GLSL::disableVertexAttribArray(h_vertTexCoords);
   GLSL::disableVertexAttribArray(h_vertPos);

   glBindBuffer(GL_ARRAY_BUFFER, 0);
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

   glActiveTexture(GL_TEXTURE0 + tex_id);
   glBindTexture(GL_TEXTURE_2D, 0);

   glUseProgram(0);
}
