#ifdef __APPLE__
#include <GLUT/glut.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif
#include <cstdlib>
#include <iostream>
#include <vector>
#include <stdio.h>
#include <algorithm>
#define GLM_FORCE_RADIANS
#include "tiny_obj_loader.h"
#include "glm/gtx/transform.hpp"
#include "GLSL.h"
#include "TextureShape.h"
#include "TexturePlane.h"
#include "MapObject.h"
#include "Image.h"
#include "Camera.h"
#include "MatrixStack.h"

#define NUM_MATERIALS 3
#define NUM_SHADERS 4
#define NUM_LIGHTS 1
#define NUM_TEXTURESHAPES 4
#define NUM_TEXTUREPLANES 6

#define BULLET_SPEED 1.0f

#define NUM_MAP_OBJECTS 18
#define NUM_WALL_BRICK 8
#define NUM_WALL_STONE 4
#define NUM_WALL_WOOD 2
#define NUM_CRATE 4

#define NUM_MAP_TARGETS 4

#define PID_BLINNPHONG 0
#define PID_SILHOUETTE 1
#define PID_CEL 2
#define PID_TEXTURE 3

#define TEXTURE_WALL_BRICK 0
#define TEXTURE_WALL_STONE 1
#define TEXTURE_WALL_WOOD 2
#define TEXTURE_CRATE 3
#define TEXTURE_GROUND_SAND 4
#define TEXTURE_SKY_TOP 5
#define TEXTURE_SKY_LEFT 6
#define TEXTURE_SKY_RIGHT 7
#define TEXTURE_SKY_FRONT 8
#define TEXTURE_SKY_BACK 9

#define SKYBOX_DIMS 400.0f
#define GND_DIMS 300.0f

using namespace std;

typedef struct Bullet {
   bool active;
   glm::vec3 direction;
   glm::vec3 dimensions;
   MapObject mobject;
} Bullet;

typedef struct GameManager {
   int time;
   glm::vec2 mouse;
   vector<TextureShape> map_shapes;
   vector<TexturePlane> map_ground_sky;
   vector<MapObject> map_objects;
   vector<MapObject> map_targets;
   bool active_targets[NUM_MAP_TARGETS];
   Bullet bullet;
} GameManager;

static GameManager gm;
Camera camera;

bool keyDown[256] = {false};
bool cull = true;
bool line = false;

vector<tinyobj::shape_t> shapeBunny;
vector<tinyobj::material_t> materialBunny;

vector<tinyobj::shape_t> shapeBullet;
vector<tinyobj::material_t> materialBullet;

GLuint bunnyPosBufID;
GLuint bunnyNorBufID;
GLuint bunnyIndBufID;

GLuint bulletPosBufID;
GLuint bulletNorBufID;
GLuint bulletIndBufID;

// GLSL handles to various parameters in the shaders

/**
 * pid[0]: Blinn Phong
 * pid[1]: Silhouette
 * pid[2]: Cel
 * pid[3]: Texture
 */
GLuint pid[NUM_SHADERS];
GLint h_ka[NUM_SHADERS];
GLint h_kd[NUM_SHADERS];
GLint h_ks[NUM_SHADERS];
GLint h_s[NUM_SHADERS];

GLint h_vertPos;
GLint h_vertNor[NUM_SHADERS];
GLint h_vertTexCoords;
GLint h_P;
GLint h_MV;
GLint h_T;
GLint h_lightPos;
GLint h_texture;

void loadGroundSky()
{
   vector<float> positionBuf;
   vector<float> texCoordsBuf;
   vector<unsigned int> indexBuf;
   
   // time
   gm.time = 0;

   // Populate the world here
   positionBuf.push_back(-1.0f);
   positionBuf.push_back(-1.0f);
   positionBuf.push_back(0.0f);
   texCoordsBuf.push_back(0.0);
   texCoordsBuf.push_back(0.0);

   positionBuf.push_back(1.0f);
   positionBuf.push_back(-1.0f);
   positionBuf.push_back(0.0f);
   texCoordsBuf.push_back(1.0);
   texCoordsBuf.push_back(0.0);

   positionBuf.push_back(-1.0f);
   positionBuf.push_back(1.0f);
   positionBuf.push_back(0.0f);
   texCoordsBuf.push_back(0.0);
   texCoordsBuf.push_back(1.0);

   positionBuf.push_back(1.0f);
   positionBuf.push_back(1.0f);
   positionBuf.push_back(0.0f);
   texCoordsBuf.push_back(1.0);
   texCoordsBuf.push_back(1.0);

   indexBuf.push_back(0);
   indexBuf.push_back(1);
   indexBuf.push_back(2);
   indexBuf.push_back(3);
   indexBuf.push_back(2);
   indexBuf.push_back(1);

   glm::mat4 GND_T = glm::translate(glm::vec3(0.0f, -12.0f, 0.0f));
   glm::mat4 GND_R = glm::rotate(-1.5707963f, glm::vec3(1.0f, 0.0f, 0.0f));
   glm::mat4 GND_S = glm::scale(glm::vec3(GND_DIMS, GND_DIMS, 1.0f));

   glm::mat4 SKY_TOP_T = glm::translate(glm::vec3(0.0f, SKYBOX_DIMS, 0.0f));
   glm::mat4 SKY_TOP_RX = glm::rotate(1.5707963f, glm::vec3(1.0f, 0.0f, 0.0f));
   glm::mat4 SKY_TOP_RY = glm::rotate(-1.5707963f, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 SKY_TOP_S = glm::scale(glm::vec3(SKYBOX_DIMS, SKYBOX_DIMS, 1.0f));

   glm::mat4 SKY_LEFT_T = glm::translate(glm::vec3(-SKYBOX_DIMS, 0.0f, 0.0f));
   glm::mat4 SKY_LEFT_R = glm::rotate(-4.7123889f, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 SKY_LEFT_S = glm::scale(glm::vec3(SKYBOX_DIMS, SKYBOX_DIMS, 1.0f));

   glm::mat4 SKY_RIGHT_T = glm::translate(glm::vec3(SKYBOX_DIMS, 0.0f, 0.0f));
   glm::mat4 SKY_RIGHT_R = glm::rotate(-1.5707963f, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 SKY_RIGHT_S = glm::scale(glm::vec3(SKYBOX_DIMS, SKYBOX_DIMS, 1.0f));
   
   glm::mat4 SKY_FRONT_T = glm::translate(glm::vec3(0.0f, 0.0f, SKYBOX_DIMS));
   glm::mat4 SKY_FRONT_R = glm::rotate(-3.14159265f, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 SKY_FRONT_S = glm::scale(glm::vec3(SKYBOX_DIMS, SKYBOX_DIMS, 1.0f));
   
   glm::mat4 SKY_BACK_T = glm::translate(glm::vec3(0.0f, 0.0f, -SKYBOX_DIMS));
   glm::mat4 SKY_BACK_R = glm::rotate(-3.14159265f, glm::vec3(0.0f, 1.0f, 0.0f));
   glm::mat4 SKY_BACK_S = glm::scale(glm::vec3(SKYBOX_DIMS, SKYBOX_DIMS, 1.0f));

   glm::mat4 Mground = GND_T * GND_R * GND_S;
   glm::mat4 Stop = SKY_TOP_T * SKY_TOP_RY * SKY_TOP_RX * SKY_TOP_S;
   glm::mat4 Sleft = SKY_LEFT_T * SKY_LEFT_R * SKY_LEFT_S;
   glm::mat4 Sright = SKY_RIGHT_T * SKY_RIGHT_R * SKY_RIGHT_S;
   glm::mat4 Sfront = SKY_FRONT_T * SKY_FRONT_R * SKY_FRONT_S;
   glm::mat4 Sback = SKY_BACK_T * SKY_BACK_S;

   glm::mat3 TTransform(1.0);
   TTransform[0][0] = 16.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 16.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;

   gm.map_ground_sky.reserve(NUM_TEXTUREPLANES);

   TexturePlane curPlane;
   
   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Mground, TTransform);
   gm.map_ground_sky.push_back(curPlane);

   TTransform[0][0] = 1.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 1.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;

   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Stop, TTransform);
   gm.map_ground_sky.push_back(curPlane);

   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Sleft, TTransform);
   gm.map_ground_sky.push_back(curPlane);

   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Sright, TTransform);
   gm.map_ground_sky.push_back(curPlane);

   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Sfront, TTransform);
   gm.map_ground_sky.push_back(curPlane);
   
   curPlane.load(positionBuf, texCoordsBuf, indexBuf, Sback, TTransform);
   gm.map_ground_sky.push_back(curPlane);
}

/**
 * map_shapes.at(0): Wall (Brick)
 * map_shapes.at(1): Wall (Stone)
 * map_shapes.at(2): Wall (Wood)
 * map_shapes.at(3): Crate
 */
void loadMap() {
   int i;
   
   glm::vec3 dimensions;
   glm::vec3 location;
   MapObject curMapObject;
   
   gm.map_shapes.reserve(NUM_TEXTURESHAPES);
   gm.map_objects.reserve(NUM_MAP_OBJECTS);
   gm.map_targets.reserve(NUM_MAP_TARGETS);
   TextureShape curShape;
   
   curShape.load("cubeTex.obj");
   
   for (i = 0; i < NUM_TEXTURESHAPES; i++) {
      gm.map_shapes.push_back(curShape);
   }

   // Stone Wall 1
   location = glm::vec3(-300.0f, 0.0f, 0.0f);
   dimensions = glm::vec3(5.0f, 50.0f, 600.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);

   // Stone Wall 2
   location = glm::vec3(300.0f, 0.0f, 0.0f);
   dimensions = glm::vec3(5.0f, 50.0f, 600.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Stone Wall 3
   location = glm::vec3(0.0f, 0.0f, 300.0f);
   dimensions = glm::vec3(600.0f, 50.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Stone Wall 4
   location = glm::vec3(0.0f, 0.0f, -300.0f);
   dimensions = glm::vec3(600.0f, 50.0f, 50.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 1
   location = glm::vec3(-200.0f, 0.0f, -150.0f);
   dimensions = glm::vec3(5.0f, 30.0f, 100.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 2
   location = glm::vec3(-200.0f, 0.0f, 150.0f);
   dimensions = glm::vec3(5.0f, 30.0f, 100.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 3
   location = glm::vec3(200.0f, 0.0f, -150.0f);
   dimensions = glm::vec3(5.0f, 30.0f, 100.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
    
   // Brick Wall 4
   location = glm::vec3(200.0f, 0.0f, 150.0f);
   dimensions = glm::vec3(5.0f, 30.0f, 100.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 5
   location = glm::vec3(-150.0f, 0.0f, -200.0f);
   dimensions = glm::vec3(100.0f, 30.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 6
   location = glm::vec3(-150.0f, 0.0f, 200.0f);
   dimensions = glm::vec3(100.0f, 30.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 7
   location = glm::vec3(150.0f, 0.0f, -200.0f);
   dimensions = glm::vec3(100.0f, 30.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Brick Wall 8
   location = glm::vec3(150.0f, 0.0f, 200.0f);
   dimensions = glm::vec3(100.0f, 30.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);

   // Wood Wall 1
   location = glm::vec3(0.0f, 0.0f, 0.0f);
   dimensions = glm::vec3(400.0f, 30.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);

   // Wood Wall 2
   location = glm::vec3(0.0f, 0.0f, 0.0f);
   dimensions = glm::vec3(5.0f, 30.0f, 400.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Crate 1
   location = glm::vec3(50.0f, -7.0f, 50.0f);
   dimensions = glm::vec3(10.0f, 10.0f, 10.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Crate 2
   location = glm::vec3(50.0f, -7.0f, -50.0f);
   dimensions = glm::vec3(10.0f, 10.0f, 10.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
 
   // Crate 3
   location = glm::vec3(-50.0f, -7.0f, 50.0f);
   dimensions = glm::vec3(10.0f, 10.0f, 10.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);
   
   // Crate 4
   location = glm::vec3(-50.0f, -7.0f, -50.0f);
   dimensions = glm::vec3(10.0f, 10.0f, 10.0f);
   curMapObject.init(dimensions, location);
   gm.map_objects.push_back(curMapObject);

   loadGroundSky();

   // Load bunny obj
   string err1 = tinyobj::LoadObj(shapeBunny, materialBunny, "bunny.obj");
   // Load bullet obj
   string err2 = tinyobj::LoadObj(shapeBullet, materialBullet, "sphere.obj");

   if (!err1.empty()) {
      cerr << err1 << endl;
   }

   if (!err2.empty()) {
      cerr << err2 << endl; 
   }

   for (i = 0; i < NUM_MAP_TARGETS; i++) {
      gm.active_targets[i] = true;
   }

   // Bunny 1
   location = glm::vec3(50.0f, -4.0f, 50.0f);
   dimensions = glm::vec3(5.0f, 5.0f, 5.0f);
   curMapObject.init(dimensions, location);
   gm.map_targets.push_back(curMapObject);

   // Bunny 2
   location = glm::vec3(-50.0f, -4.0f, 50.0f);
   dimensions = glm::vec3(6.0f, 6.0f, 6.0f);
   curMapObject.init(dimensions, location);
   gm.map_targets.push_back(curMapObject);

   // Bunny 3
   location = glm::vec3(50.0f, -4.0f, -50.0f);
   dimensions = glm::vec3(4.0f, 4.0f, 4.0f);
   curMapObject.init(dimensions, location);
   gm.map_targets.push_back(curMapObject);

   // Bunny 4
   location = glm::vec3(-50.0f, -4.0f, -50.0f);
   dimensions = glm::vec3(4.0f, 4.0f, 4.0f);
   curMapObject.init(dimensions, location);
   gm.map_targets.push_back(curMapObject);

   // Bullet
   gm.bullet.active = false;
   gm.bullet.dimensions = glm::vec3(0.25f, 0.25f, 0.5f);
   gm.bullet.mobject.init(dimensions, glm::vec3(0.0f, 0.0f, 0.0f));
}

void initGL()
{
   int i;
   
   //////////////////////////////////////////////////////
   // Initialize GL for the whole scene
   //////////////////////////////////////////////////////
   
   // Set background color
   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
   // Enable z-buffer test
   glEnable(GL_DEPTH_TEST);

   for (i = 0; i < NUM_TEXTURESHAPES; i++) {
      gm.map_shapes.at(i).init();
   }

   for (i = 0; i < NUM_TEXTUREPLANES; i++) {
      gm.map_ground_sky.at(i).init();
   }

   // Initialze GL for Bunny
   const vector<float> &posBufBunny = shapeBunny[0].mesh.positions;
	glGenBuffers(1, &bunnyPosBufID);
	glBindBuffer(GL_ARRAY_BUFFER, bunnyPosBufID);
	glBufferData(GL_ARRAY_BUFFER, posBufBunny.size()*sizeof(float), 
    &posBufBunny[0], GL_STATIC_DRAW);
	
	// Send the normal array to the GPU
	const vector<float> &norBufBunny = shapeBunny[0].mesh.normals;
	glGenBuffers(1, &bunnyNorBufID);
	glBindBuffer(GL_ARRAY_BUFFER, bunnyNorBufID);
	glBufferData(GL_ARRAY_BUFFER, norBufBunny.size()*sizeof(float), 
    &norBufBunny[0], GL_STATIC_DRAW);
	
	// Send the index array to the GPU
	const vector<unsigned int> &indBufBunny = shapeBunny[0].mesh.indices;
	glGenBuffers(1, &bunnyIndBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bunnyIndBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
    indBufBunny.size()*sizeof(unsigned int), &indBufBunny[0], GL_STATIC_DRAW);

	// Unbind the arrays
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

   
   // Initialze GL for Bullet
   const vector<float> &posBufBullet = shapeBullet[0].mesh.positions;
	glGenBuffers(1, &bulletPosBufID);
	glBindBuffer(GL_ARRAY_BUFFER, bulletPosBufID);
	glBufferData(GL_ARRAY_BUFFER, posBufBullet.size()*sizeof(float), 
    &posBufBullet[0], GL_STATIC_DRAW);
	
	// Send the normal array to the GPU
	const vector<float> &norBufBullet = shapeBullet[0].mesh.normals;
	glGenBuffers(1, &bulletNorBufID);
	glBindBuffer(GL_ARRAY_BUFFER, bulletNorBufID);
	glBufferData(GL_ARRAY_BUFFER, norBufBullet.size()*sizeof(float), 
    &norBufBullet[0], GL_STATIC_DRAW);
	
	// Send the index array to the GPU
	const vector<unsigned int> &indBufBullet = shapeBullet[0].mesh.indices;
	glGenBuffers(1, &bulletIndBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bulletIndBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
    indBufBullet.size()*sizeof(unsigned int), &indBufBullet[0], GL_STATIC_DRAW);

	// Unbind the arrays
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

   //////////////////////////////////////////////////////
   // Intialize the shaders
   //////////////////////////////////////////////////////
   
   // Create shader handles
   string vShaderName = "vert.glsl";
   string fBlinnPhongShader = "frag_blinnphong.glsl";
   string fSilhouetteShader = "frag_silhouette.glsl";
   string fCelShader = "frag_cel.glsl";
   string fTextureShader = "frag_texture.glsl";
   GLint rc;
   GLuint VS = glCreateShader(GL_VERTEX_SHADER);
   GLuint FS[NUM_SHADERS];

   for (i = 0; i < NUM_SHADERS; i++) {
      FS[i] = glCreateShader(GL_FRAGMENT_SHADER);
   }
   
   // Read shader sources
   const char *vshader = GLSL::textFileRead(vShaderName.c_str());
   const char *fshader[NUM_SHADERS];

   fshader[PID_BLINNPHONG] = GLSL::textFileRead(fBlinnPhongShader.c_str());
   fshader[PID_SILHOUETTE] = GLSL::textFileRead(fSilhouetteShader.c_str());
   fshader[PID_CEL] = GLSL::textFileRead(fCelShader.c_str());
   fshader[PID_TEXTURE] = GLSL::textFileRead(fTextureShader.c_str());
   
   glShaderSource(VS, 1, &vshader, NULL);

   for (i = 0; i < NUM_SHADERS; i++) {
      glShaderSource(FS[i], 1, &fshader[i], NULL);
   }

   // Compile vertex shader
   glCompileShader(VS);
   GLSL::printError();
   glGetShaderiv(VS, GL_COMPILE_STATUS, &rc);
   GLSL::printShaderInfoLog(VS);
   if(!rc) {
      printf("Error compiling vertex shader %s\n", vShaderName.c_str());
   }
   
   // Compile fragment shaders
   for (i = 0; i < NUM_SHADERS; i++) {
      glCompileShader(FS[i]);
      GLSL::printError();
      glGetShaderiv(FS[i], GL_COMPILE_STATUS, &rc);
      GLSL::printShaderInfoLog(FS[i]);
      if (!rc) {
         printf("Error compiling fragment shader %s\n", fshader[i]);
      }
   }
   
   // Create shader programs and link
   for (i = 0; i < NUM_SHADERS; i++) {
      pid[i] = glCreateProgram();
      glAttachShader(pid[i], VS);
      glAttachShader(pid[i], FS[i]);
      glLinkProgram(pid[i]);
      GLSL::printError();
      glGetProgramiv(pid[i], GL_LINK_STATUS, &rc);
      GLSL::printProgramInfoLog(pid[i]);
      if (!rc) {
         printf("Error linking shaders %s and %s\n", vShaderName.c_str(), 
          fshader[i]);
      }
      
      h_vertPos = GLSL::getAttribLocation(pid[i], "vertPos");
      h_vertNor[i] = GLSL::getAttribLocation(pid[i], "vertNor");
      h_P = GLSL::getUniformLocation(pid[i], "P");
      h_MV = GLSL::getUniformLocation(pid[i], "MV");

      if (i == PID_BLINNPHONG || i == PID_CEL) { 
         h_lightPos = GLSL::getUniformLocation(pid[i], "lightPos");
         h_ka[i] = GLSL::getUniformLocation(pid[i], "ka");
         h_kd[i] = GLSL::getUniformLocation(pid[i], "kd");
         h_ks[i] = GLSL::getUniformLocation(pid[i], "ks");
         h_s[i] = GLSL::getUniformLocation(pid[i], "s");
      }

      if (i == PID_TEXTURE) { 
         h_vertTexCoords = GLSL::getAttribLocation(pid[i], "vertTexCoords");
         h_texture = GLSL::getUniformLocation(pid[i], "texture");
         h_T = GLSL::getUniformLocation(pid[i], "T");
      }
   }

   //////////////////////////////////////////////////////
   // Initialize textures
   //
   // map_shapes.at(0): Wall (Brick)
   // map_shapes.at(1): Wall (Stone)
   // map_shapes.at(2): Wall (Wood)
   // map_shapes.at(3): Crate
   //
   // map_ground_sky.at(0): Ground (Sand)
   // map_ground_sky.at(1): Sky (Top)
   // map_ground_sky.at(2): Sky (Left)
   // map_ground_sky.at(3): Sky (Right)
   // map_ground_sky.at(4): Sky (Front)
   // map_ground_sky.at(5): Sky (Back)
   //////////////////////////////////////////////////////

   // id, bmp filename, clamp (0 == repeat)
   gm.map_shapes.at(0).texInit(0, "resources/wall_brick.bmp", 0);
   gm.map_shapes.at(1).texInit(1, "resources/wall_stone.bmp", 0);
   gm.map_shapes.at(2).texInit(2, "resources/wall_wood.bmp", 0);
   gm.map_shapes.at(3).texInit(3, "resources/crate.bmp", 1);

   gm.map_ground_sky.at(0).texInit(4, "resources/ground_sand.bmp", 0);
   gm.map_ground_sky.at(1).texInit(5, "resources/sky_top.bmp", 1);
   gm.map_ground_sky.at(2).texInit(6, "resources/sky_left.bmp", 1);
   gm.map_ground_sky.at(3).texInit(7, "resources/sky_right.bmp", 1);
   gm.map_ground_sky.at(4).texInit(8, "resources/sky_front.bmp", 1);
   gm.map_ground_sky.at(5).texInit(9, "resources/sky_back.bmp", 1);
   
   // Check GLSL
   GLSL::checkVersion();
}

void reshapeGL(int w, int h)
{
   // Set view size
   glViewport(0, 0, (GLsizei)w, (GLsizei)h);
   // Set camera aspect ratio
   camera.setWindowSize(w, h);
}

void drawGL()
{
   int i;
   int nIndices = (int)shapeBunny[0].mesh.indices.size();
   glm::mat3 TTransform(1.0);
   
   // Clear buffers
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // Enable backface culling
   if(cull) {
      glEnable(GL_CULL_FACE);
   } else {
      glDisable(GL_CULL_FACE);
   }
   if(line) {
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
   } else {
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   }
   
   // Create matrix stacks
   MatrixStack P, MV;
   // Apply camera transforms
   P.pushMatrix();
   camera.applyProjectionMatrix(&P);
   MV.pushMatrix();
   camera.applyViewMatrix(&MV);

   for (i = 0; i < NUM_TEXTUREPLANES; i++) {
      gm.map_ground_sky.at(i).draw(&MV, &P, h_MV, h_P, h_T, pid[PID_TEXTURE], 
       h_vertPos, h_vertTexCoords, h_texture);
   }
   
   // Draw stone walls (4)
   glUseProgram(pid[PID_TEXTURE]);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
   
   TTransform[0][0] = 8.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 2.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;
   glUniformMatrix3fv( h_T, 1, GL_TRUE, glm::value_ptr(TTransform));

   for (i = 0; i < NUM_WALL_STONE; i++) {
      MV.pushMatrix();
      MV.translate(gm.map_objects.at(i).getLocation());
      MV.scale(gm.map_objects.at(i).getDimensions());
      glUniformMatrix4fv( h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
      gm.map_shapes.at(TEXTURE_WALL_STONE).draw(h_vertPos, 
       h_vertNor[PID_TEXTURE], h_vertTexCoords, h_texture);
      MV.popMatrix();
   }
   
   glUseProgram(0);

   // Draw brick walls (8)
   glUseProgram(pid[PID_TEXTURE]);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
   
   TTransform[0][0] = 2.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 1.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;
   glUniformMatrix3fv( h_T, 1, GL_TRUE, glm::value_ptr(TTransform));

   for (; i < NUM_WALL_STONE + NUM_WALL_BRICK; i++) {
      MV.pushMatrix();
      MV.translate(gm.map_objects.at(i).getLocation());
      MV.scale(gm.map_objects.at(i).getDimensions());
      glUniformMatrix4fv( h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
      gm.map_shapes.at(TEXTURE_WALL_BRICK).draw(h_vertPos, h_vertNor[PID_TEXTURE], 
       h_vertTexCoords, h_texture);
      MV.popMatrix();
   }
  
   glUseProgram(0);

   // Draw wood walls (2)
   glUseProgram(pid[PID_TEXTURE]);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
   
   TTransform[0][0] = 2.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 1.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;
   glUniformMatrix3fv( h_T, 1, GL_TRUE, glm::value_ptr(TTransform));

   for (; i < NUM_WALL_STONE + NUM_WALL_BRICK + NUM_WALL_WOOD; i++) {
      MV.pushMatrix();
      MV.translate(gm.map_objects.at(i).getLocation());
      MV.scale(gm.map_objects.at(i).getDimensions());
      glUniformMatrix4fv( h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
      gm.map_shapes.at(TEXTURE_WALL_WOOD).draw(h_vertPos, h_vertNor[PID_TEXTURE], 
       h_vertTexCoords, h_texture);
      MV.popMatrix();
   }
  
   glUseProgram(0);

   // Draw crates (4)
   glUseProgram(pid[PID_TEXTURE]);
   glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
   
   TTransform[0][0] = 1.0f;
   TTransform[0][1] = 0.0f;
   TTransform[0][2] = 0.0f;
   TTransform[1][0] = 0.0f;
   TTransform[1][1] = 1.0f;
   TTransform[1][2] = 0.0f;
   TTransform[2][0] = 0.0f;
   TTransform[2][1] = 0.0f;
   TTransform[2][2] = 1.0f;
   glUniformMatrix3fv( h_T, 1, GL_TRUE, glm::value_ptr(TTransform));

   for (; i < NUM_WALL_STONE + NUM_WALL_BRICK + NUM_WALL_WOOD + NUM_CRATE; i++) {
      MV.pushMatrix();
      MV.translate(gm.map_objects.at(i).getLocation());
      MV.scale(gm.map_objects.at(i).getDimensions());
      glUniformMatrix4fv( h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
      gm.map_shapes.at(TEXTURE_CRATE).draw(h_vertPos, h_vertNor[PID_TEXTURE], 
       h_vertTexCoords, h_texture);
      MV.popMatrix();
   }
  
   glUseProgram(0);

   // Bunny 1 (Blinn Phong)
   if (gm.active_targets[0]) {
      glUseProgram(pid[PID_BLINNPHONG]);
      glUniform3fv(h_lightPos, 1, glm::value_ptr(glm::vec3(100.0f, 100.0f, 100.0f)));
      glUniform3fv(h_ka[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.2f, 0.2f, 0.2f)));
      glUniform3fv(h_kd[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.3f, 0.9f, 0.4f)));
      glUniform3fv(h_ks[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(1.0f, 0.9f, 0.8f)));
      glUniform1f(h_s[PID_BLINNPHONG], 200.0f);

      // Send projection matrix to the GPU
      glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));

      // Enable and bind position array
      GLSL::enableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyPosBufID);
      glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

      GLSL::enableVertexAttribArray(h_vertNor[PID_BLINNPHONG]);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyNorBufID);
      glVertexAttribPointer(h_vertNor[PID_BLINNPHONG], 3, GL_FLOAT, GL_FALSE, 0, 0);

      // Bind index array
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bunnyIndBufID);

      // Apply model matrix and send to GPU
      MV.pushMatrix();
      MV.translate(gm.map_targets.at(0).getLocation());
      MV.scale(gm.map_targets.at(0).getDimensions());
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));

      // Draw
      glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
      MV.popMatrix();

      // Disable and unbind plane
      GLSL::disableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

      // Unbind the program
      glUseProgram(0);
   }

   // Bunny 2 (Silhouette)
   if (gm.active_targets[1]) {
      glUseProgram(pid[PID_SILHOUETTE]);

      // Send projection matrix to the GPU
      glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));

      // Enable and bind position array
      GLSL::enableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyPosBufID);
      glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

      GLSL::enableVertexAttribArray(h_vertNor[PID_SILHOUETTE]);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyNorBufID);
      glVertexAttribPointer(h_vertNor[PID_SILHOUETTE], 3, GL_FLOAT, GL_FALSE, 0, 0);

      // Bind index array
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bunnyIndBufID);

      // Apply model matrix and send to GPU
      MV.pushMatrix();
      MV.translate(gm.map_targets.at(1).getLocation());
      MV.scale(gm.map_targets.at(1).getDimensions());
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));

      // Draw
      glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
      MV.popMatrix();

      // Disable and unbind plane
      GLSL::disableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

      // Unbind the program
      glUseProgram(0);
   }
   
   // Bunny 3 (Cel Shaded)
   if (gm.active_targets[2]) {
      glUseProgram(pid[PID_CEL]);
      glUniform3fv(h_lightPos, 1, glm::value_ptr(glm::vec3(100.0f, 100.0f, 100.0f)));
      glUniform3fv(h_ka[PID_CEL], 1, glm::value_ptr(glm::vec3(0.1f, 0.1f, 0.1f)));
      glUniform3fv(h_kd[PID_CEL], 1, glm::value_ptr(glm::vec3(0.4f, 0.1f, 0.9f)));
      glUniform3fv(h_ks[PID_CEL], 1, glm::value_ptr(glm::vec3(0.8f, 1.0f, 0.6f)));
      glUniform1f(h_s[PID_CEL], 100.0f);

      // Send projection matrix to the GPU
      glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));

      // Enable and bind position array
      GLSL::enableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyPosBufID);
      glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

      GLSL::enableVertexAttribArray(h_vertNor[PID_CEL]);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyNorBufID);
      glVertexAttribPointer(h_vertNor[PID_CEL], 3, GL_FLOAT, GL_FALSE, 0, 0);

      // Bind index array
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bunnyIndBufID);

      // Apply model matrix and send to GPU
      MV.pushMatrix();
      MV.translate(gm.map_targets.at(2).getLocation());
      MV.scale(gm.map_targets.at(2).getDimensions());
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));

      // Draw
      glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
      MV.popMatrix();

      // Disable and unbind plane
      GLSL::disableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

      // Unbind the program
      glUseProgram(0);
   }

   // Bunny 4 (Blinn Phong 2)
   if (gm.active_targets[3]) {
      glUseProgram(pid[PID_BLINNPHONG]);
      glUniform3fv(h_lightPos, 1, glm::value_ptr(glm::vec3(100.0f, 100.0f, 100.0f)));
      glUniform3fv(h_ka[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.2f, 0.2f, 0.2f)));
      glUniform3fv(h_kd[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.9f, 0.9f, 0.2f)));
      glUniform3fv(h_ks[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.5f, 0.9f, 0.5f)));
      glUniform1f(h_s[PID_BLINNPHONG], 400.0f);

      // Send projection matrix to the GPU
      glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));

      // Enable and bind position array
      GLSL::enableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyPosBufID);
      glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

      GLSL::enableVertexAttribArray(h_vertNor[PID_BLINNPHONG]);
      glBindBuffer(GL_ARRAY_BUFFER, bunnyNorBufID);
      glVertexAttribPointer(h_vertNor[PID_BLINNPHONG], 3, GL_FLOAT, GL_FALSE, 0, 0);

      // Bind index array
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bunnyIndBufID);

      // Apply model matrix and send to GPU
      MV.pushMatrix();
      MV.translate(gm.map_targets.at(3).getLocation());
      MV.scale(gm.map_targets.at(3).getDimensions());
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));

      // Draw
      glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
      MV.popMatrix();

      // Disable and unbind plane
      GLSL::disableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

      // Unbind the program
      glUseProgram(0);
   }

   // Bullet
   if (gm.bullet.active) {
      glUseProgram(pid[PID_BLINNPHONG]);
      glUniform3fv(h_lightPos, 1, glm::value_ptr(glm::vec3(100.0f, 100.0f, 100.0f)));
      glUniform3fv(h_ka[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.0f, 0.0f, 0.0f)));
      glUniform3fv(h_kd[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.0f, 0.0f, 0.0f)));
      glUniform3fv(h_ks[PID_BLINNPHONG], 1, glm::value_ptr(glm::vec3(0.1f, 0.1f, 0.1f)));
      glUniform1f(h_s[PID_BLINNPHONG], 400.0f);
   
      // Send projection matrix to the GPU
      glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr(P.topMatrix()));
   
      // Enable and bind position array
      GLSL::enableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, bulletPosBufID);
      glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

      GLSL::enableVertexAttribArray(h_vertNor[PID_BLINNPHONG]);
      glBindBuffer(GL_ARRAY_BUFFER, bulletNorBufID);
      glVertexAttribPointer(h_vertNor[PID_BLINNPHONG], 3, GL_FLOAT, GL_FALSE, 0, 0);

      // Bind index array
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bulletIndBufID);

      // Apply model matrix and send to GPU
      MV.pushMatrix();
      MV.translate(gm.bullet.mobject.getLocation());
      MV.scale(gm.bullet.mobject.getDimensions());
      glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
   
      // Draw
      nIndices = (int)shapeBullet[0].mesh.indices.size();
      glDrawElements(GL_TRIANGLES, nIndices, GL_UNSIGNED_INT, 0);
      MV.popMatrix();
   
      // Disable and unbind plane
      GLSL::disableVertexAttribArray(h_vertPos);
      glBindBuffer(GL_ARRAY_BUFFER, 0);
      glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

      // Unbind the program
      glUseProgram(0);
   }

   // Pop stacks
   MV.popMatrix();
   P.popMatrix();

   // Double buffer
   glutSwapBuffers();
}

void mouseGL(int button, int state, int x, int y)
{
   if (button == GLUT_LEFT_BUTTON && !gm.bullet.active) {
      glm::vec3 position = camera.getPosition();
      
      gm.bullet.active = true;
      gm.bullet.direction = camera.getDirection();
      gm.bullet.mobject.init(gm.bullet.dimensions, position);
   }
}

void passiveMotionGL(int x, int y)
{
   gm.mouse.x = x;
   gm.mouse.y = y;
}

void keyboardGL(unsigned char key, int x, int y)
{
   keyDown[key] = true;
   switch(key) {
      case 27:
         // ESCAPE
         exit(0);
         break;
      case 'c':
         cull = !cull;
         break;
      case 'l':
         line = !line;
         break;
   }
}

void keyboardUpGL(unsigned char key, int x, int y)
{
   keyDown[key] = false;
}

void idleGL()
{
   int curtime = glutGet(GLUT_ELAPSED_TIME);
   int dt = curtime - gm.time;
   int i;
   bool noCollisions = true;
   MatrixStack OBJ;
   glm::mat4 bulletLocation;
   glm::vec3 curLocation;
   Contacts curCollision;

   // Update every 60Hz
   if(dt > 1000.0/60.0) {
      gm.time = curtime;
      camera.update(keyDown, gm.mouse, gm.map_objects);
      glutPostRedisplay();
   }

   // Update every 240Hz
   if (dt > 1000.0/240.0) {
      if (gm.bullet.active) {
         curLocation = gm.bullet.mobject.getLocation();
         curLocation += BULLET_SPEED * gm.bullet.direction;
         gm.bullet.mobject.init(gm.bullet.dimensions, curLocation);

         OBJ.pushMatrix();
         OBJ.translate(curLocation);
         bulletLocation = OBJ.topMatrix();
         OBJ.popMatrix();

         // Check collisions with non-targets
         for (i = 0; i < gm.map_objects.size() && noCollisions; i++) {
            curCollision = BoxCollision(bulletLocation, gm.bullet.dimensions,
             gm.map_objects.at(i).getMatrix(), 
             gm.map_objects.at(i).getDimensions());
            
            if (curCollision.positions.size() != 0) {
               noCollisions = false;
               gm.bullet.active = false;
            }
         }

         // Check collisions with targets
         for (i = 0; i < gm.map_targets.size() && noCollisions; i++) {
            curCollision = BoxCollision(bulletLocation, glm::vec3(5.0f, 5.0f, 5.0f),
             gm.map_targets.at(i).getMatrix(), glm::vec3(5.0f, 5.0f, 5.0f));

            if (curCollision.positions.size() != 0) {
               noCollisions = false;
               gm.bullet.active = false;
               gm.active_targets[i] = false;
            }
         }
      }
   }
}

int main(int argc, char **argv)
{
   glutInit(&argc, argv);
   glutInitWindowSize(WINDOW_SIZE_X, WINDOW_SIZE_Y);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
   glutCreateWindow("Zach Mintzer");
   glutMouseFunc(mouseGL);
   glutPassiveMotionFunc(passiveMotionGL);
   glutKeyboardFunc(keyboardGL);
   glutKeyboardUpFunc(keyboardUpGL);
   glutReshapeFunc(reshapeGL);
   glutDisplayFunc(drawGL);
   glutIdleFunc(idleGL);
   glutFullScreen();
   glutSetCursor(GLUT_CURSOR_CROSSHAIR);
   loadMap();
   initGL();
   glutMainLoop();
   return 0;
}
